@@ -1,6 +1,7 @@
 /*
 
   Copyright (C) 2006-2018 Michael Goffioul <michael.goffioul@swing.be>
+  Copyright (C) 2019-2020 John Donoghue <john.donoghue@ieee.org>
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -24,6 +25,8 @@
 #include <octave/defun-dld.h>
 #include <octave/Cell.h>
 #include <octave/parse.h>
+#include <octave/oct-map.h>
+#include <octave/lo-sysdep.h>
 
 #ifdef HAVE_CONFIG_H
 #  include "config.h"
@@ -42,6 +45,9 @@
 static std::wstring
 string_to_wstring (const std::string& s)
 {
+#if HAVE_OCTAVE_U8_TO_WSTRING
+  return octave::sys::u8_to_wstring (s);
+#else
   const char* cs = s.c_str ();
   int len = s.length ();
 
@@ -52,11 +58,15 @@ string_to_wstring (const std::string& s)
   std::use_facet<std::ctype<wchar_t> >(std::locale ()).widen (
     cs, cs+len, &tmp[0]);
   return std::wstring (&tmp[0], len);
+#endif
 }
 
 static std::string
 wstring_to_string (const std::wstring& ws)
 {
+#if HAVE_OCTAVE_U8_TO_WSTRING
+  return octave::sys::u8_from_wstring (ws);
+#else
   const wchar_t* wcs = ws.c_str ();
   int len = ws.length ();
 
@@ -67,9 +77,30 @@ wstring_to_string (const std::wstring& ws)
   std::use_facet<std::ctype<wchar_t> >(std::locale ()).narrow (
       wcs, wcs+len, ' ', &tmp[0]);
   return std::string (&tmp[0], len);
+#endif
 }
 
-class octave_com_object : public octave_base_value, public octave_auto_shlib
+static std::string
+hresult_to_string (HRESULT hr)
+{
+  wchar_t errstring[100+1];
+  std::string str;
+
+  if (FormatMessageW (
+    FORMAT_MESSAGE_FROM_SYSTEM,
+    0, hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), errstring,
+    100, 0) == 0)
+    {
+      str = "Unknown";
+    }
+  else
+    {
+      str = wstring_to_string(errstring);
+    }
+  return str;
+}
+
+class octave_com_object : public octave_base_value
 {
 public:
   // Constructors
@@ -281,11 +312,8 @@ octave_com_object::subsasgn (const std::string& type, const std::list<octave_val
         ovl (1) = (idx.front ()) (0);
         ovl (2) = rhs;
         OCTAVE__FEVAL ("com_set", ovl, 0);
-        if (!error_state)
-          {
-            count++;
-            retval = octave_value (this);
-          }
+        count++;
+        retval = octave_value (this);
       }
     else if (type.length () > 2 && type[1] == '(')
       {
@@ -295,33 +323,27 @@ octave_com_object::subsasgn (const std::string& type, const std::list<octave_val
         new_idx.push_back (*it++);
         new_idx.push_back (*it++);
         octave_value_list u = subsref (type.substr (0, 2), new_idx, 1);
-        if (!error_state)
+        if (u.length())
           {
             std::list<octave_value_list> next_idx (idx);
             next_idx.erase (next_idx.begin ());
             next_idx.erase (next_idx.begin ());
             u (0).subsasgn (type.substr (2), next_idx, rhs);
-            if (!error_state)
-              {
-                count++;
-                retval = octave_value (this);
-              }
+            count++;
+            retval = octave_value (this);
           }
       }
     else if (type.length () > 1 && type[1] == '.')
       {
         // get property and continue assignment
         octave_value_list u = subsref (type.substr (0, 1), idx, 1);
-        if (!error_state)
+        if (u.length())
           {
             std::list<octave_value_list> next_idx (idx);
             next_idx.erase (next_idx.begin ());
             u (0).subsasgn(type.substr (1), next_idx, rhs);
-            if (!error_state)
-              {
-                count++;
-                retval = octave_value (this);
-              }
+            count++;
+            retval = octave_value (this);
           }
       }
     else
@@ -367,6 +389,78 @@ terminate_com()
 }
 #endif
 
+static void
+feature(const std::string &name, octave_value &val, int dir)
+{
+  static octave_scalar_map features;
+
+  if (features.nfields() == 0)
+    {
+      features.assign("COM_SafeArraySingleDim", octave_value(0));
+    }
+
+  if (dir == 0)
+    {
+      features.setfield(name, val);
+    }
+  else
+    {
+      val = features.getfield(name);
+    }
+}
+
+static octave_value
+get_feature(const std::string &name)
+{
+  octave_value v;
+  feature(name, v, 1);
+  return v;
+}
+
+// PKG_ADD: autoload ("windows_feature", which ("__COM__"));
+DEFUN_DLD(windows_feature, args, , 
+          "-*- texinfo -*-\n \
+@deftypefn {Loadable Function} {} windows_feature (@var{name})\n \
+@deftypefnx {Loadable Function} {} windows_feature (@var{name}, @var{value})\n \
+Set or get a feature value.\n \
+\n \
+@var{name} - name of feature to get or set.@*\n \
+@var{value} - value to set for feature.@*\n \
+@end deftypefn")
+{
+  octave_value_list retval;
+
+  if (args.length () != 1 && args.length () != 2)
+    {
+      error ("windows_feature: Expected a name and optional value");
+      return retval;
+    }
+
+  std::string name;
+  if (! args(0).is_string())
+    {
+      error ("windows_feature: Expected a string for first argument");
+      return retval;
+    }
+
+  name = args(0).string_value ();
+
+  if (args.length () == 1)
+    {
+      octave_value val;
+      feature (name, val, 1);
+      retval(0) = val;
+    }
+  else
+    {
+      octave_value val = args(1);
+      feature (name, val, 0);
+    }
+
+  return retval;
+}
+
+
 // PKG_ADD: autoload ("com_atexit", which ("__COM__"));
 // xPKG_ADD: #atexit ("com_atexit");
 DEFUN_DLD(com_atexit, args, , 
@@ -446,6 +540,30 @@ destroy (app); \n \
 }
 
 #ifdef USING_WINDOWS
+
+template<typename T>
+octave_value 
+safearray_to_octave_value(const dim_vector& dv, SAFEARRAY *parray)
+{
+  T *pvalues;
+  HRESULT hr;
+  Array<T> cell(dv);
+
+  hr = SafeArrayAccessData(parray, (void **)&pvalues);
+  if (FAILED(hr))
+  {
+    warning("failed accessing array data");
+    return octave_value(Matrix()); /*return invalid data */
+  }
+
+  for (int k = 0; k < cell.numel(); k++) {
+    cell(k) = pvalues[k];
+  }
+  SafeArrayUnaccessData(parray);
+
+  return octave_value(cell);
+}
+
 static octave_value
 com_to_octave (VARIANT *var)
 {
@@ -454,17 +572,41 @@ com_to_octave (VARIANT *var)
   switch (var->vt)
   {
   case VT_DISPATCH:
-    retval = octave_value (new octave_com_object (var->pdispVal, true));
+    if (var->pdispVal)
+        retval = octave_value (new octave_com_object (var->pdispVal, true));
+    else
+        retval = octave_value (Matrix());
     break;
   case VT_BOOL:
     retval = octave_value (var->boolVal == VARIANT_TRUE ? true : false);
     break;
+  case VT_I1:
+    retval = octave_value (var->cVal);
+    break;
+  case VT_UI1:
+    retval = octave_value (var->bVal);
+    break;
+  case VT_I2:
+    retval = octave_value (var->iVal);
+    break;
+  case VT_UI2:
+    retval = octave_value (var->uiVal);
+    break;
   case VT_I4:
     retval = octave_value (var->lVal);
     break;
+  case VT_UI4:
+    retval = octave_value (var->ulVal);
+    break;
+  case VT_I8:
+    retval = octave_value (var->llVal);
+    break;
   case VT_INT:
     retval = octave_value (var->intVal);
     break;
+  case VT_UINT:
+    retval = octave_value (var->uintVal);
+    break;
   case VT_BSTR:
     retval = octave_value (wstring_to_string (std::wstring (var->bstrVal)));
     break;
@@ -507,6 +649,45 @@ com_to_octave (VARIANT *var)
 
         switch (subtype)
         {
+        case VT_BSTR:
+          {
+            VARIANT *pvar;
+            HRESULT hr;
+            BSTR cellVal;
+            long arrDims[2] = {0,0};
+            Cell cell (dv);
+
+            hr=SafeArrayAccessData (arr, (void **)&pvar);
+
+            if (FAILED (hr))
+              {
+                warning("failed accessing array data");
+                retval = octave_value (Matrix ()); // return invalid data
+                break;
+              }
+            else
+              {
+                for (int k=0; k<cell.numel (); k++)
+                  {
+                    arrDims[0] = k;
+                    if (SafeArrayGetElement (arr,arrDims,&cellVal)==S_OK)
+                      {
+                       // retval=octave_value (cellVal); // retval used as temporary value
+                       cell (k) = octave_value (wstring_to_string (std::wstring ( cellVal)));
+                      }
+                    else 
+                      {
+                        warning ("error accessing value (%d)", k);
+                        // returns whatever data we already have
+                        break;
+                      }
+                  }
+
+                retval = octave_value (cell); // return invalid data
+              }
+
+          }
+          break;
         case VT_VARIANT:
           {
             VARIANT *pvar;
@@ -532,140 +713,28 @@ com_to_octave (VARIANT *var)
           }
           break;
         case VT_R4:
-          {
-            VARIANT *pvar;
-            HRESULT hr;
-            FLOAT cellVal;
-            long arrDims[2] = {0,0};
-            Array<float> cell (dv);
-
-            hr=SafeArrayAccessData (arr, (void **)&pvar);
-            if (FAILED (hr))
-              {
-                warning("failed accessing array data");
-                retval = octave_value (Matrix ()); // return invalid data
-                break;
-              }
-            else
-              {
-                for (int dimNum=0; dimNum<dimCount; dimNum++)
-                  {
-                    arrDims[0] = dimNum; // overwritten for 1-dim arrays
-
-                    for (int k=0; k<dv (dimCount-1); k++) // index last value of array (handles 1- or 2-dimensions)
-                      {
-                        arrDims[dimCount-1] = k;
-                        if (SafeArrayGetElement (arr,arrDims,&cellVal)==S_OK)
-                          {
-                            retval=octave_value (cellVal); // retval used as temporary value
-                            if (dimCount==2) // 2D array
-                              cell (dv (0)*k+dimNum)=retval.float_value ();
-                            else // 1D array
-                              cell (k)=retval.float_value ();
-                          }
-                        else 
-                          {
-                            warning ("error accessing value (%d,%d)", dimNum,k);
-                            // returns whatever data we already have
-                            break;
-                          }
-                      }
-                  }
-                SafeArrayUnaccessData (arr);
-              }
-            retval = octave_value (cell);
-          }
+          retval = safearray_to_octave_value<float>(dv, arr);
+          break;
+        case VT_R8:
+          retval = safearray_to_octave_value<double>(dv, arr);
+          break;
+        case VT_I1:
+          retval = safearray_to_octave_value<octave_int8>(dv, arr);
           break;
         case VT_UI1:
-          {
-            VARIANT *pvar;
-            HRESULT hr;
-            BYTE cellVal;
-            long arrDims[2] = {0,0};
-            Array<octave_uint8> cell (dv);
-
-            hr=SafeArrayAccessData (arr, (void **)&pvar);
-            if (FAILED (hr))
-              {
-                warning("failed accessing array data");
-                retval = octave_value(Matrix()); // return invalid data
-                break;
-              }
-            else
-              {
-                for (int dimNum=0; dimNum<dimCount; dimNum++)
-                  {
-                    arrDims[0] = dimNum; // overwritten for 1-dim arrays
-
-                    for (int k=0; k<dv(dimCount-1); k++) // index last value of array (handles 1- or 2-dimensions)
-                      {
-                        arrDims[dimCount-1] = k;
-                        if (SafeArrayGetElement (arr, arrDims, &cellVal)==S_OK)
-                          {
-                            retval=octave_value (cellVal); // retval used as temporary variable
-                            if (dimCount==2) // 2D array
-                              cell (dv (0)*k+dimNum)=retval.ushort_value ();
-                            else // 1D array
-                              cell (k) = retval.ushort_value ();
-                          }
-                        else
-                          {
-                            warning ("error accessing value (%d,%d)", dimNum, k);
-                            // returns whatever data we already have
-                            break;
-                          }
-                      }
-                  }
-                SafeArrayUnaccessData (arr);
-              }
-            retval = octave_value (cell);
-          }
+          retval = safearray_to_octave_value<octave_uint8>(dv, arr);
           break;
         case VT_I2:
-          {
-            VARIANT *pvar;
-            HRESULT hr;
-            SHORT cellVal;
-            long arrDims[2] = {0,0};
-            Array<octave_int16> cell (dv);
-
-            //warning("%d-D VT_I2 array detected",dimCount);
-            hr=SafeArrayAccessData (arr, (void **)&pvar);
-            if (FAILED (hr))
-              {
-                warning ("failed accessing array data");
-                retval = octave_value (Matrix ()); // return invalid data
-                break;
-              }
-            else
-              {
-                for (int dimNum=0; dimNum<dimCount; dimNum++)
-                  {
-                    arrDims[0] = dimNum; // overwritten for 1-dim arrays
-
-                    for (int k=0; k<dv (dimCount-1); k++) // index last value of array (handles 1- or 2-dimensions)
-                      {
-                        arrDims[dimCount-1] = k;
-                        if (SafeArrayGetElement (arr,arrDims,&cellVal)==S_OK)
-                          {
-                            retval=octave_value (cellVal); // retval used as temporary variable
-                            if(dimCount==2) // 2D array
-                              cell (dv (0)*k+dimNum) = retval.int_value ();
-                            else // 1D array
-                              cell (k) = retval.int_value ();
-                          }
-                        else
-                          {
-                            warning ("error accessing value (%d,%d)", dimNum, k);
-                            // returns whatever data we already have
-                           break;
-                          }
-                      }
-                  }
-                SafeArrayUnaccessData (arr);
-              }
-            retval = octave_value (cell);
-          }
+          retval = safearray_to_octave_value<octave_int16>(dv, arr);
+          break;
+        case VT_UI2:
+          retval = safearray_to_octave_value<octave_uint16>(dv, arr);
+          break;
+        case VT_I4:
+          retval = safearray_to_octave_value<octave_int32>(dv, arr);
+          break;
+        case VT_UI4:
+          retval = safearray_to_octave_value<octave_uint32>(dv, arr);
           break;
         default:
           {
@@ -687,15 +756,22 @@ com_to_octave (VARIANT *var)
 }
 
 static SAFEARRAY*
-make_safearray_from_dims(const dim_vector& dv, VARTYPE vt)
+make_safearray_from_dims(const dim_vector& dv, VARTYPE vt, int maxDim)
 {
   SAFEARRAY *arr;
   SAFEARRAYBOUND *bounds;
 
-  bounds = (SAFEARRAYBOUND*)LocalAlloc (LMEM_FIXED|LMEM_ZEROINIT, sizeof (SAFEARRAYBOUND)*dv.length ());
-  for (int k=0; k<dv.length (); k++)
+  int arrayDim;
+  if (dv.length () > maxDim && maxDim != 0) 
+    arrayDim = maxDim;
+  else
+    arrayDim = dv.length ();
+	
+  bounds = (SAFEARRAYBOUND*)LocalAlloc (LMEM_FIXED|LMEM_ZEROINIT, sizeof (SAFEARRAYBOUND)*arrayDim);
+  for (int k=0; k<arrayDim; k++)
     bounds[k].cElements = dv (k);
-  arr = SafeArrayCreate (vt, dv.length (), bounds);
+
+  arr = SafeArrayCreate (vt, arrayDim, bounds);
   LocalFree (bounds);
 
   return arr;
@@ -706,6 +782,10 @@ octave_to_com(const octave_value& ov, VARIANT *var)
 {
   VariantInit (var);
 
+  int maxDims = 0;
+  if(get_feature("COM_SafeArraySingleDim").int_value() == 1)
+    maxDims = 1;
+
   if (ov.is_string ())
   {
     var->vt = VT_BSTR;
@@ -728,7 +808,8 @@ octave_to_com(const octave_value& ov, VARIANT *var)
   else if (ov.is_real_matrix ())
   {
     NDArray M = ov.array_value ();
-    SAFEARRAY *arr = make_safearray_from_dims (M.dims (), VT_R8);
+
+    SAFEARRAY *arr = make_safearray_from_dims (M.dims (), VT_R8, maxDims);
     double *data;
 
     SafeArrayAccessData (arr, (void**)&data);
@@ -742,7 +823,7 @@ octave_to_com(const octave_value& ov, VARIANT *var)
   else if (ov.OV_ISCELL ())
   {
     Cell M = ov.cell_value ();
-    SAFEARRAY *arr = make_safearray_from_dims (M.dims (), VT_VARIANT);
+    SAFEARRAY *arr = make_safearray_from_dims (M.dims (), VT_VARIANT, maxDims);
     VARIANT *data;
 
     SafeArrayAccessData (arr, (void**)&data);
@@ -777,7 +858,7 @@ do_invoke (const char *fname, WORD flag, const octave_value_list& args)
 
   // Get method dispatch ID
   method_name = string_to_wstring (args (1).string_value ());
-  if (error_state)
+  if (method_name.length() == 0)
     {
       error ("%s: invalid property/method name as argument 2", fname);
       return retval;
@@ -806,7 +887,8 @@ do_invoke (const char *fname, WORD flag, const octave_value_list& args)
   if ((hr = com->com_iface ()->Invoke (method_ID, IID_NULL, LOCALE_USER_DEFAULT, flag, &dispParams,
                                        &result, NULL, NULL)) != S_OK)
     {
-      error ("%s: property/method invocation on the COM object failed with error `0x%08x'", fname, hr);
+      std::string errstr = hresult_to_string(hr);
+      error ("%s: property/method invocation on the COM object failed with error `0x%08x' - %s", fname, hr, errstr.c_str());
       goto cleanup;
     }
 
@@ -886,7 +968,7 @@ Call get function on COM object @var{obj}. Returns any result in @var{S}\n \
   if (args.length () == 1)
     retval = octave_value (Cell (do_invoke_list ("com_get", DISPATCH_PROPERTYGET, OV_COMOBJ (args (0)))));
   else
-    retval = do_invoke ("com_get", DISPATCH_PROPERTYGET, args);
+    retval = do_invoke ("com_get", DISPATCH_PROPERTYGET|DISPATCH_METHOD, args);
 #endif
   return retval;
 }
@@ -1014,3 +1096,65 @@ DEFUN_DLD(__COM__, args, , "internal function")
 #endif
   return retval;
 }
+
+// PKG_ADD: autoload ("__windows_pkg_lock__", which ("__COM__"));
+// PKG_ADD: __windows_pkg_lock__(1);
+// PKG_DEL: __windows_pkg_lock__(0);
+#ifdef DEFMETHOD_DLD
+DEFMETHOD_DLD (__windows_pkg_lock__, interp, args, , "internal function")
+{
+  octave_value retval;
+  if (args.length () >= 1)
+    {
+      if (args(0).int_value () == 1)
+        interp.mlock();
+      else if (args(0).int_value () == 0 &&  interp.mislocked("__windows_pkg_lock__"))
+        interp.munlock("__windows_pkg_lock__");
+    }
+  return retval;
+}
+#else
+DEFUN_DLD(__windows_pkg_lock__, args, ,  "internal function")
+{
+  octave_value retval;
+  return retval;
+}
+#endif
+
+#if 0
+%!testif HAVE_WINDOWS_H
+%! wshell = actxserver ("WScript.Shell");
+%! assert (!isempty (wshell));
+#! assert (strcmp (wshell.CurrentDirectory, pwd));
+
+%!testif HAVE_WINDOWS_H
+%! wshell = actxserver ("WScript.Shell");
+%! # type REG_SZ
+%! val = wshell.RegRead('HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\CurrentVersion');
+%! assert (isa (val, 'char'));
+%! # type REG_DWORD
+%! val = wshell.RegRead('HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\InstallDate');
+%! assert (isscalar (val));
+%! # type REG_BINARY
+%! val = wshell.RegRead('HKLM\SOFTWARE\Microsoft\DirectX\InstalledVersion');
+%! assert (iscell (val));
+%! assert (length (val) > 0);
+%! # type REG_EXPANDSZ
+%! val = wshell.RegRead('HKLM\SOFTWARE\Microsoft\Windows\CurrentVersion\ProgramFilesPath');
+%! assert (isa (val, 'char'));
+
+%!testif HAVE_WINDOWS_H
+%! fail ("actxserver(0)", "invalid ActiveX server name");
+
+%!test
+%! assert(windows_feature("COM_SafeArraySingleDim"), 0);
+%! windows_feature("COM_SafeArraySingleDim", 1);
+%! assert(windows_feature("COM_SafeArraySingleDim"), 1);
+%! windows_feature("COM_SafeArraySingleDim", 0);
+
+%!test
+%! fail ("windows_feature()", "Expected a name and optional value");
+%! fail ("windows_feature(1)", "Expected a string");
+%! fail ("windows_feature('a', 1, 1)", "Expected a name and optional value");
+
+#endif
